/*                                               -*- C -*- */
/**
 *  @file  wrapper.c
 *  @brief The wrapper adapts the interface of OpenTURNS and of the wrapped code
 *
 *  (C) Copyright 2005-2007 EDF-EADS-Phimeca
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  @author: $LastChangedBy: dutka $
 *  @date:   $LastChangedDate: 2008-06-26 13:50:17 +0200 (jeu, 26 jun 2008) $
 *  Id:      $Id: wrapper.c 862 2008-06-26 11:50:17Z dutka $
 */

#include <stdio.h>
#include "WrapperInterface.h"
#include "WrapperCommon.h"

/* WARNING : Please read the following lines
 *
 * In this program, we make the assumption that the end user wishes to
 * call a function (aka NumericalMathFunction) named "wcode" (wcode stands
 * for wrapped code).
 * In order to individualize the wrapper to the user's needs, we encourage
 * you (as the developer of this wrapper) to renamed any occurence of "wcode"
 * (in either case) to the real name of the function.
 * It will also avoid any confusion with other "wcode"s written by other entities
 * or developers.
 */


/* If you plan to link this wrapper against a FORTRAN library,
 * remember to change the name WCODE for your actual FORTRAN subroutine name.
 * Otherwise you can ignore these lines.
 *
 * Remember that FORTRAN passes its arguments by reference, not by value as C and C++
 * usually do. So you need to pass the pointer to the arguments rather than its value.
 * This is true for single values (integers, reals, etc.) but not for arrays that are
 * already pointers in the C/C++ environment. Those ones can directly be passed as
 * "values" though they are pointers indeed.
 * Be careful that C and C++ arrays start from 0 and FORTRAN from 1!
 * Be also very careful with the size of the value your plan to pass. Integers in C/C++
 * are not INTEGER*8 in many cases. Float or doubles do not necessarily match REAL*4
 * or REAL*8 in FORTRAN.
 *
 * FORTRAN gives no clue for preventing const values from being altered. So you need to
 * protect them by copying them before calling the FORTRAN subroutine if this import
 * to you.
 *
 * Summary: there are only exceptions to the rule and you need to
 * know exactly what you are doing! You may be desappointed at first, but it will keep
 * you away from segmentation violation and other similar fantasies. ;-)
 */
#define WCODE_F77 F77_FUNC(wcode,WCODE)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void WCODE_F77(double * I1, double * I2, double * I3, double * O1, double * O2);


/* NOTE: all the 'dbg_' function are for debugging needs */


/*
 *  This is the declaration of function named 'wcode' into the wrapper.
 */
  

#ifdef __cplusplus
extern "C" {
#endif

  /*
*********************************************************************************
*                                                                               *
*                             wcode function                                    *
*                                                                               *
*********************************************************************************
*/


  /* The wrapper information informs the NumericalMathFunction object that loads the wrapper of the
   * signatures of the wrapper functions. In particular, it hold the size of the input NumericalPoint
   * (inSize_) and of the output NumericalPoint (outSize_).
   * Those information are also used by the gradient and hessian functions to set the correct size
   * of the returned matrix and tensor.
   */

  /* The getInfo function is mandatory */
  enum WrapperErrorCode func_getInfo_wcode(void * p_state, struct WrapperInformation * p_info)
  {
    dbg_printEntrance("func_getInfo_wcode");
    dbg_printState("func_getInfo_wcode", p_state);

    struct WrapperExchangedData * p_exchangedData = p_state;

    p_info->inSize_  = getNumberOfVariables(p_exchangedData, WRAPPER_IN);
    p_info->outSize_ = getNumberOfVariables(p_exchangedData, WRAPPER_OUT);

    dbg_printExit("func_getInfo_wcode");

    return WRAPPER_OK;
  }


  /* The state creation/deletion functions allow the wrapper to create or delete a memory location
   * that it will manage itself. It can save in this location any information it needs. The OpenTURNS
   * platform only ensures that the wrapper will receive the state (= the memory location) it works
   * with. If many wrappers are working simultaneously or if the same wrapper is called concurrently,
   * this mechanism will avoid any collision or confusion.
   * The consequence is that NO STATIC DATA should be used in the wrapper OR THE WRAPPER WILL BREAKE
   * one day. You may think that you can't do without static data, but in general this is the footprint
   * of a poor design. But if you persist to use static data, do your work correctly and make use
   * of mutex (for instance) to protect your data against concurrent access. But don't complain about
   * poor computational performance! 
   */

  /* The createState function is optional */
  enum WrapperErrorCode func_createState_wcode(void ** p_p_state, const struct WrapperExchangedData * p_exchangedData)
  {
    dbg_printEntrance("func_createState_wcode");
    dbg_printWrapperExchangedData("func_createState_wcode", p_exchangedData);

    /* Here we copy the data read in the WML description file into the state */
    copyWrapperExchangedData( (struct WrapperExchangedData **) p_p_state, p_exchangedData);

    dbg_printState("func_createState_wcode", *p_p_state);
    dbg_printExit("func_createState_wcode");

    return WRAPPER_OK;
  }

  /* The deleteState function is optional */
  enum WrapperErrorCode func_deleteState_wcode(void * p_state)
  {
    dbg_printEntrance("func_deleteState_wcode");
    dbg_printState("func_deleteState_wcode", p_state);

    /* We free the memory allocated for the state*/
    freeWrapperExchangedData( (struct WrapperExchangedData *) p_state);

    dbg_printExit("func_deleteState_wcode");

    return WRAPPER_OK;
  }






  /* Any function declared into the wrapper may declare three actual function prefixed with
   * 'init_', 'exec_' and 'finalize_' folowed by the name of the function, here 'wcode'.
   *
   * The 'init_' function is only called once when the NumericalMathFunction object is created.
   * It allows the wrapper to set some internal state, read some external file, prepare the function
   * to run, etc. It takes only one argument, the internal state as created by the 
   *
   * The 'exec_' function is intended to execute what the wrapper is done for: compute an mathematical
   * function or anything else. It takes the internal state pointer as its first argument, the input
   * NumericalPoint pointer as the second and the output NumericalPoint pointer as the third.
   *
   * The 'finalize_' function is only called once when the NumericalMathFunction object is destroyed.
   * It allows the wrapper to flush anything before unloading.
   *
   * Only the 'exec_' function is mandatory.
   */


  /**
   * Initialization function
   * This function is called once just before the wrapper first called to initialize
   * it, ie create a temparary subdirectory (remember that the wrapper may be called
   * concurrently), store exchanged data in some internal repository, do some
   * pre-computational operation, etc.
   */
  enum WrapperErrorCode func_init_wcode(void * p_state)
  {
    dbg_printEntrance("func_init_wcode");
    dbg_printState("func_init_wcode", p_state);

    /* Do something if you need to initialize the wrapper or its environment.
     * If not, you can remove the entire function.
     */

    dbg_printExit("func_init_wcode");

   return WRAPPER_OK;
  }




  /**
   * Execution function
   * This function is called by the platform to do the real work of the wrapper. It may be
   * called concurrently, so be aware of not using shared or global data not protected by
   * a critical section.
   * This function has a mathematical meaning. It operates on one vector (aka point) and
   * returns another vector.
   */
  enum WrapperErrorCode func_exec_wcode(void * p_state, const struct point * inPoint, struct point * outPoint)
  {
    dbg_printEntrance("func_exec_wcode");
    dbg_printState("func_exec_wcode", p_state);
    dbg_printPoint("func_exec_wcode", inPoint);




    /* The real computation is here */

    /* Call to the FORTRAN subroutine named WCODE.
     * We have made the assumption that the inputs values won't be touched by the call,
     * so we pass the arguments directly without any protection (copy).
     */
    WCODE_F77(&(inPoint->data_[0])  /* I1 */ ,
	      &(inPoint->data_[1])  /* I2 */ ,
	      &(inPoint->data_[2])  /* I3 */ ,
	      &(outPoint->data_[0]) /* O1 */ ,
	      &(outPoint->data_[0]) /* O2 */ );



    dbg_printPoint("func_exec_wcode", outPoint);
    dbg_printExit("func_exec_wcode");

    return WRAPPER_OK;
  }




  /**
   * Finalization function
   * This function is called once just before the wrapper is unloaded. It is the place to flush
   * any output file or free any allocated memory. When this function returns, the wrapper is supposed
   * to have all its work done, so it is not possible to get anymore information from it after that.
   */
  enum WrapperErrorCode func_finalize_wcode(void * p_state)
  {
    dbg_printEntrance("func_finalize_wcode");
    dbg_printState("func_finalize_wcode", p_state);

    /* Do something if you need to clean the wrapper or its environment.
     * If not, you can remove the entire function.
     */

    dbg_printExit("func_finalize_wcode");

    return WRAPPER_OK;
  }




#ifdef __cplusplus
} /* end extern "C" */
#endif
